# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
package Storage::FtpFileStorage;

use strict;
use bigint;

use Storage::Storage;
use Logging;
use AgentConfig;
use HelpFuncs;

use POSIX;
use IPC::Run;
use Symbol;
use Error;
use File::Copy;

use Storage::Splitter;
use Storage::Counter;
use Net::FTP;

use vars qw|@ISA|;

@ISA = qw|Storage::Storage|;

$Error::Debug = 1;

sub _init {
  my ($self, %options) = @_;

  $self->SUPER::_init(%options);
  $self->{passive_mode} = $options{passive_mode};
  $self->{management_node} = 1;

  if ($options{output_dir} =~ /^(ftp|ftps):\/\/(?:([^:]*)(?::([^:]*))?@)?([^\/:@]+)(?::([0-9]+))?\/(.*?)$/) {

    my %ftp;

    $ftp{'protocol'} = $1;
    $ftp{'login'} = defined $2 ? $2 : '';
    $ftp{'password'} = defined $3 ? $3 : '';

    if ($ftp{'password'} eq '' && defined $ENV{'FTP_PASSWORD'}) {
      $ftp{'password'} = $ENV{'FTP_PASSWORD'};
    }

    if ($ftp{'login'} eq '') {
      $ftp{'login'} = 'anonymous';
      $ftp{'password'} = 'plesk@plesk.com' if ($ftp{'password'} eq '');
    }

    $ftp{'server'} = $4;
    $ftp{'port'} = $5;
    $ftp{'path'} = $6;
          
    $self->{ftpsettings} = \%ftp;
    
    my %ftpOptions;
    my $ftpConnection;
    
    Logging::debug("Trying to connect to $ftp{'server'} in " . $self->{passive_mode} ? "passive" : "active" . " mode...");
    
    if ($ftp{'protocol'} eq 'ftps') {
      my $ftpsModuleAvailable = 0;
      eval {
        require Net::FTPSSL;
        Net::FTPSSL->import();
        $ftpsModuleAvailable = 1;
      };
      if (!$ftpsModuleAvailable) {
        Logging::error("Server environment doea not support FTP over SSL. Net::FTPSSL module must be installed to turn this feature on");
        return 1;
      }
      $ftpConnection = Net::FTPSSL->new($ftp{'server'}, Port => $ftp{'port'}, Encryption => 'E', Debug => 1);
    } else {
      $ftpOptions{Debug} = 1;
      $ftpOptions{Passive} = $self->{passive_mode};
      $ftpConnection = Net::FTP->new($ftp{'server'}, %ftpOptions);
    }

    if (! defined $ftpConnection ) {
      Logging::error("Could not connect to $ftp{'server'}", 'FtpError');
      return 1;
    }

    if (!$ftpConnection->login($ftp{'login'}, $ftp{'password'})) {
      Logging::error( "FTP: Can't login to the ftp server",'FtpError' );
      return 1;
    }

    if ($ftp{'path'} ne "" && !$ftpConnection->cwd($ftp{'path'})) {
      Logging::error( "FTP: Can't change dir to $ftp{path}",'FtpError' );
      return 1;
    }
    
    $ftpConnection->binary;
    
    $self->{ftpconnection} = $ftpConnection;
    Logging::debug("Connected successfully.");
  }

  Logging::debug("-" x 60);
  Logging::debug("FTP FILE storage initialized.");
  Logging::debug("Backup node: $self->{ftpsettings}->{server}");
  Logging::debug("Mode: " . $self->{passive_mode} ? "passive" : "active");
  Logging::debug("Base directory: $self->{ftpsettings}->{path}");
  Logging::debug("Space reserved: $self->{space_reserved}");
  Logging::debug("Gzip bundles: " . ($self->{gzip_bundle} ? "yes" : "no"));
  Logging::debug("Bundle split size: " . ($self->{split_size} || "do not split"));  
  Logging::debug("-" x 60);
  $self->reserveSpace();
}

sub setRemoteExecution {
    my ($self) = @_;
    $self->{management_node} = 0;
}

sub createRepositoryIndex{
  my ( $self, $index ) = @_;
  my @filesToUpload;
  if( $index ){
    Logging::debug("Create repository index: $index");
    my $destDir = $self->getFullOutputPath()."/.discovered";
    system("mkdir", "-p", "$destDir") if not -e $destDir;
    open INDEXFILE, "> $destDir/$index";
    close INDEXFILE;
    my @fileToUpload = ($index, 0);
    push @filesToUpload, \@fileToUpload;
    $self->uploadFilesToFtp( $destDir, \@filesToUpload );    
  }
}

sub writeDiscovered{
  my $self = shift;

  my ($destDir, $files) = @{$self->_writeDiscovered(@_)};
  $self->uploadFilesToFtp($destDir, $files);
}

sub moveFileToDiscovered {
  my ($self, $srcPath, $newName, $dumpDirPath, $dumpXmlName) = @_;
  
  $destDir = $self->_moveFileToDiscovered($srcPath, $newName, $dumpDirPath, $dumpXmlName);
  $self->uploadFilesToFtp($destDir, [[$newName]]);
}

sub addTar {
  my $self = shift;

  my ($ret, $destDir, $files) = @{$self->_addTar(@_)};
  $self->uploadFilesToFtp($destDir, $files);
  return $ret;
}

sub uploadFileToFtp {
  my ($self, $sourcePath, $destPath) = @_;
  if (! $self->{ftpconnection}->put($sourcePath, $destPath)) {
    Error::throw Error::Simple("Can't upload $sourcePath to $destPath: " . $self->{ftpconnection}->message);
  }
}

sub uploadFilesToFtp {
  my ($self, $destDir, $files) = @_;

# todo check connection, reconnect  or several attempts to upload

  if ( index( $destDir, $self->getFullOutputPath() )==0 ) {
    $destDir = substr( $destDir, length( $self->getFullOutputPath() ) + 1 ) ;
  }  
  
  if( index( $destDir, -1, 1 ) eq '/' ) {
    $destDir = substr( $destDir, 0, length($destDir)-1 );
  }
  
  $self->{ftpconnection}->mkdir($destDir, 1);
 
  for my $file( @{$files} ){
    Logging::debug("PUT file: " . $self->getFullOutputPath() . '/' . "$destDir/$file->[0]" . " TO PATH: " . "$destDir/$file->[0]");
    my $destinationOnFtp;
    if ($destDir eq '') {
        $destinationOnFtp = $file->[0];
    }else {
        $destinationOnFtp = "$destDir/$file->[0]";
    }
    $self->uploadFileToFtp($self->getFullOutputPath() . '/' . "$destDir/$file->[0]", $destinationOnFtp);    
  }
  
  for my $file( @{$files} ){
    unlink($self->getFullOutputPath() . '/' . "$destDir/$file->[0]") or Logging::error("Cannot delete temp file ". $self->getFullOutputPath() . "/" . $destDir ."/".$file->[0]);
  }  
}

sub CleanupFiles()
{
  my $self = shift;

  $self->_waitChild();

  my $path = $self->getFullOutputPath();
  my @files = $self->getDumpFiles();
  foreach my $file(@files ){
     if (-e $path."/".$file) {
       $self->uploadFileToFtp($path."/".$file, $file);
     }
  }

  $self->_removeFiles();
}

1;

# Local Variables:
# mode: cperl
# cperl-indent-level: 2
# indent-tabs-mode: nil
# tab-width: 4
# End:
