# Copyright 1999-2017. Plesk International GmbH. All rights reserved.

import sys
import logging as log
from optparse import OptionParser, SUPPRESS_HELP
from fsmng_stat import FileStat
from fsmng_helpers import PermsHelper, FsError, FileNotInPackageError, PackageNotFoundError

# FIXME
# PPP-32877 Check postfix package permissions fails on centos6
# PPP-32767 Fsmng produces and error for psa-vpn
EXCLUDED_PACKAGES = ['postfix', 'psa-vpn']


def set_ac(path, opts):
    helper = PermsHelper()
    if any([opts.with_perms, opts.with_owner, opts.with_group, opts.with_type]):
        helper.what_to_fix(with_perms=opts.with_perms, with_owner=opts.with_owner,
                           with_group=opts.with_group, with_type=opts.with_type)
    fstat = None
    if any([opts.perms, opts.owner, opts.group, opts.filetype]):
        if opts.filetype == FileStat.LNK and not opts.symlink_to:
            raise ValueError("--symlink-to should be specified for --filetype=symlink")
        fstat = FileStat(perms=opts.perms, user=opts.owner, group=opts.group, typ=opts.filetype, link_to=opts.symlink_to)
    helper.fix_file(path, fstat)


def check_ac(path, opts):
    helper = PermsHelper()
    if any([opts.with_perms, opts.with_owner, opts.with_group, opts.with_type]):
        helper.what_to_fix(with_perms=opts.with_perms, with_owner=opts.with_owner,
                           with_group=opts.with_group, with_type=opts.with_type)
    fstat = None
    if any([opts.perms, opts.owner, opts.group, opts.filetype]):
        if opts.filetype == FileStat.LNK and not opts.symlink_to:
            raise ValueError("--symlink-to should be specified for --filetype=symlink")
        fstat = FileStat(perms=opts.perms, user=opts.owner, group=opts.group, typ=opts.filetype, link_to=opts.symlink_to)
    helper.check_file(path, fstat)


def fix_package_ac(pkg_name):
    helper = PermsHelper(pkg=pkg_name)
    helper.fix_package()


def check_package_ac(pkg_name):
    helper = PermsHelper(pkg=pkg_name)
    helper.check_package()


def fix_plesk_packages():
    pkgs = PermsHelper.get_plesk_packages()
    for pkg in pkgs:
        if pkg in EXCLUDED_PACKAGES:
            log.info("Skip fixing package '%s': it is in exclude list", pkg)
            continue
        log.info("Fixing package '%s'", pkg)
        try:
            pkg_fixer = PermsHelper(pkg=pkg)
            pkg_fixer.fix_package()
        except FileNotInPackageError as e:
            log.warning(e)
        except PackageNotFoundError as e:
            log.warning("Skip fixing '%s': %s", pkg, e)


def check_plesk_packages():
    pkgs = PermsHelper.get_plesk_packages()
    errors = []
    for pkg in pkgs:
        if pkg in EXCLUDED_PACKAGES:
            log.info("Skip checking package '%s': it is in exclude list", pkg)
            continue
        log.info("Checking package '%s'", pkg)
        try:
            pkg_fixer = PermsHelper(pkg=pkg)
            pkg_fixer.check_package()
        except FsError as e:
            errors += str(e).split("\n")
        except FileNotInPackageError as e:
            log.warning(e)
        except PackageNotFoundError as e:
            log.warning("Skip checking '%s': %s", pkg, e)
    if errors:
        raise FsError("\n".join(errors))


def do_main(argv):
    parser = OptionParser(usage="Usage: %prog --check-ac|--set-ac <path>"
                                " [--perms <permission>,<permission>]"
                                " [--owner <owner>,<owner>]"
                                " [--group <group>,<group>]")

    parser.add_option("-v", "--verbose", action="count", help="Increase verbosity (can be specified multiple times).")

    parser.add_option("-c", "--check-ac", default=None,
                      help="checks file permissions and ownership")
    parser.add_option("-s", "--set-ac", default=None,
                      help="sets file permissions and ownership")

    parser.add_option("--with-perms", default=False, action="store_true", help=SUPPRESS_HELP)
    parser.add_option("--with-owner", default=False, action="store_true", help=SUPPRESS_HELP)
    parser.add_option("--with-group", default=False, action="store_true", help=SUPPRESS_HELP)
    parser.add_option("--with-type", default=False, action="store_true", help=SUPPRESS_HELP)

    parser.add_option("-p", "--perms", default=None, help="file permissions")
    parser.add_option("-o", "--owner", default=None, help="file owner")
    parser.add_option("-g", "--group", default=None, help="file group")
    parser.add_option("-t", "--filetype", default=None, help="file type")
    parser.add_option("-l", "--symlink-to", default=None, help="symlink destination")

    parser.add_option("--check-plesk-packages", default=False, action="store_true", dest="check_plesk",
                      help="Check permissions/ownership of all Plesk packages")
    parser.add_option("--fix-plesk-packages", default=False, action="store_true", dest="fix_plesk",
                      help="Fix permissions/ownership of all Plesk packages")
    parser.add_option("--check-package", default=None, metavar="PACKAGE", help="Check permissions of all files in PACKAGE")
    parser.add_option("--fix-package", default=None, metavar="PACKAGE", help="Fix permissions of all files in PACKAGE")

    (options, args) = parser.parse_args()

    log_level = log.WARNING
    if options.verbose == 1:
        log_level = log.INFO
    elif options.verbose >= 2:
        log_level = log.DEBUG
    log.basicConfig(format='%(levelname)s:%(message)s', level=log_level, stream=sys.stderr)

    if options.check_ac:
        check_ac(options.check_ac, options)
    elif options.set_ac:
        set_ac(options.set_ac, options)
    elif options.check_plesk:
        check_plesk_packages()
    elif options.fix_plesk:
        fix_plesk_packages()
    elif options.check_package:
        check_package_ac(options.check_package)
    elif options.fix_package:
        fix_package_ac(options.fix_package)
    else:
        raise ValueError("Neither check-ac nor set-ac options are specified.")


def main():
    try:
        do_main(sys.argv)
    except FsError as ex:
        sys.stderr.write("%s\n" % ex)
        sys.exit(0)
    except Exception as ex:
        log.error('%s', ex)
        log.debug(ex, exc_info=True)
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
