# Copyright 1999-2017. Plesk International GmbH. All rights reserved.

import os
import glob
import pwd
import grp
import re
import logging as log
import plesk_subprocess as subprocess
from fsmng_permissions import FILES_AC
from fsmng_stat import FileStat

RPM_BIN = "/bin/rpm"
DPKG_BIN = "/usr/bin/dpkg"
DPKG_QUERY_BIN = "/usr/bin/dpkg-query"
APT_CACHE_DIR = "/var/cache/apt/archives"
APT_GET_BIN = "/usr/bin/apt-get"

PLESK_PACKAGER = 'Plesk <info@plesk.com>'
PARALLELS_PACKAGER = 'Parallels <info@parallels.com>'


class FsError(Exception):
    pass


class FileNotInPackageError(Exception):
    pass


class PackageNotFoundError(Exception):
    pass


def get_owners(names):
    """ Turn comma-separated sysuser names into pair: (list of uids, list of names). """
    uids = []
    users = []
    for name in names.split(','):
        try:
            uids.append(pwd.getpwnam(name).pw_uid)
            users.append(name)
        except KeyError:
            log.warning("User '%s' not found. Skipping.", name)
    return uids, users


def get_groups(names):
    """ Turn comma-separated sysgroup names into pair: (list of gids, list of names). """
    gids = []
    groups = []
    for name in names.split(','):
        try:
            gids.append(grp.getgrnam(name).gr_gid)
            groups.append(name)
        except KeyError:
            log.warning("Group '%s' not found. Skipping.", name)
    return gids, groups


def validate_perms(path, st, perms=None):
    """ Validate permissions on path.
        If permissions are incorrent then return string description of error.
        Otherwise return None.
    """
    if not perms:
        return None
    actual = oct(st.st_mode & 07777)
    expected = perms.split(',')
    if actual not in expected:
        return "Incorrect permissions on %s." \
               " Expected: one of %s. Actual: %s." \
               % (path, perms, actual)
    return None


def validate_owner(path, st, owner=None):
    """ Validate owner of path.
        If owner is incorrent then return string description of error.
        Otherwise return None.
    """
    if not owner:
        return None
    actual = st.st_uid
    (expected_uids, expected_names) = get_owners(owner)
    if actual not in expected_uids:
        try:
            actual_name = pwd.getpwuid(actual).pw_name
        except KeyError:
            actual_name = actual
        return "Incorrect owner of %s." \
               " Expected: %s (%s). Actual: %s (%s)." \
               % (path, owner, ','.join(str(x) for x in expected_uids),
                  actual_name, actual)
    return None


def validate_group(path, st, group=None):
    """ Validate group of path.
        If group is incorrent then return string description of error.
        Otherwise return None.
    """
    if not group:
        return None
    actual = st.st_gid
    (expected_gids, expected_names) = get_groups(group)
    if actual not in expected_gids:
        try:
            actual_name = grp.getgrgid(actual).gr_name
        except KeyError:
            actual_name = actual
        return "Incorrect group of %s." \
               " Expected: %s (%s). Actual: %s (%s)." \
               % (path, group, ','.join(str(x) for x in expected_gids),
                  actual_name, actual)
    return None


def validate_type(path, st, filetype=None, link_to=None):
    if filetype is None:
        return None
    act = FileStat(mode=st.st_mode)
    if act.typ != filetype:
        return "Incorrect type of %s. Expected: %s. Actual: %s" % (path, filetype, act.typ)
    if filetype == FileStat.LNK:
        act_link_to = os.readlink(path)
        if act_link_to != link_to:
            return "Symlink %s points to incorrect place. Expected: %s. Actual: %s" \
                   % (path, link_to, act_link_to)
    return None


def check_call_with_logging(cmd):
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = popen.communicate()
    if popen.returncode == 0:
        log.debug("Running %s", cmd)
        log.debug("Stdout: %s", out)
        log.debug("Stderr: %s", err)
    else:
        log.info("Running %s", cmd)
        log.info("Stdout: %s", out)
        log.info("Stderr: %s", err)
        raise subprocess.CalledProcessError(returncode=popen.returncode, cmd=" ".join(cmd))


class RpmFilePerms:
    """ Helper to acquire proper file permissions by inspecting corresponding rpm-package meta-info """
    _pkg_name = None
    _permissions = None
    _file_list = None

    def __init__(self, pkg_name=None):
        self._pkg_name = pkg_name

    @staticmethod
    def is_installed():
        """ Check if rpm is installed. """
        return os.path.exists(RPM_BIN)

    def get_file_perms(self, path):
        """ Get FileStat object for path as described in corresponding package
            or raise FileNotInPackageError.
        """
        npath = os.path.normpath(path)
        if self._permissions:
            if npath in self._permissions:
                return self._permissions[npath]
            raise FileNotInPackageError("File permissions are not found for %s" % path)

        self._permissions = {}

        cmd = [RPM_BIN, "-qf", path, "--dump"]
        if self._pkg_name:
            cmd = [RPM_BIN, "-q", self._pkg_name, "--dump"]
        content_data = subprocess.check_output(cmd)
        for line in content_data.split("\n"):
            if not line:
                continue
            try:
                fs = self._parse_rpm_dump_line(line)
                self._permissions[line.split()[0]] = fs
            except ValueError as e:
                log.warning("failed to parse '%s' : %s. Skipping.", line, e)

        if npath in self._permissions:
            return self._permissions[npath]
        raise FileNotInPackageError("File permissions are not found for %s" % path)

    @staticmethod
    def _parse_rpm_dump_line(line):
        fields = line.split()
        fs = FileStat(mode=fields[4], user=fields[5], group=fields[6])
        if fs.typ == FileStat.LNK:
            fs.link_to = fields[10]
        return fs

    @staticmethod
    def get_plesk_packages():
        cmd = [RPM_BIN, '-qa', '--queryformat', '%{NAME}\t%{PACKAGER}\n']
        pkgs = []
        content_data = subprocess.check_output(cmd)
        for line in content_data.split("\n"):
            if line.endswith("\t" + PLESK_PACKAGER) or line.endswith("\t" + PARALLELS_PACKAGER):
                pkgs.append(line.split("\t")[0])
        return pkgs

    def package_files_list(self):
        if not self._pkg_name:
            raise Exception("Package name is not set")
        if self._file_list:
            return self._file_list

        cmd = [RPM_BIN, '-ql', self._pkg_name]
        raw_list = subprocess.check_output(cmd).split("\n")
        # filter out lines like `(contains no files)`
        self._file_list = filter(lambda x: x and x.startswith('/'), raw_list)
        return self._file_list


class DebFilePerms:
    """ Helper to acquire proper file's permissions by inspecting corresponding deb-package """
    PACKAGE_PATTERN = "Package: "
    VERSION_PATTERN = "Version: "

    _pkg_name = None
    _pkg_version = None
    _pkg_path = None
    _permissions = None
    _file_list = None

    _apt_get_updated = False

    def __init__(self, pkg_name=None):
        self._pkg_name = pkg_name

    @staticmethod
    def is_installed():
        """ Check if Apt/dpkg stuff is installed. """
        return os.path.exists(DPKG_BIN)

    def get_file_perms(self, path):
        """ Get tuple (perms, owner, group) for path
            as described in correspongind package or raise FileNotInPackageError.
        """
        try:
            pkg_name = self._pkg_name
            if not pkg_name:
                cmd = [DPKG_BIN, "-S", path]
                # output format:
                # package1[, package2 ...]: path
                output = subprocess.check_output(cmd)
                pkg_name = output[:output.index(":")].split(", ")[0]
            pkg_version = self._installed_dpkg_package_version(pkg_name)
            log.debug("File %s is from %s (%s)" % (path, pkg_name, pkg_version))
            pkg_path = self._get_package_path(pkg_name, pkg_version)
            return self._get_file_perms_from_package(path, pkg_path)
        except subprocess.CalledProcessError as e:
            raise FileNotInPackageError("%s" % e)

    def _installed_dpkg_package_version(self, pkg_name):
        if pkg_name == self._pkg_name and self._pkg_version:
            log.debug("return cached version of package %s: %s", pkg_name, self._pkg_version)
            return self._pkg_version
        log.debug("detecting version of package %s", pkg_name)
        cmd = [DPKG_BIN, "-s", pkg_name]
        output = subprocess.check_output(cmd)
        for line in output.split("\n"):
            if line.startswith(self.VERSION_PATTERN):
                version = line[len(self.VERSION_PATTERN):]
                if pkg_name == self._pkg_name:
                    self._pkg_version = version
                return version
        raise PackageNotFoundError("failed to detect version of package %s" % pkg_name)

    def _get_file_perms_from_package(self, file_path, pkg_path, content_data=None):
        pat = os.path.normpath(file_path)
        if self._permissions:
            if pat in self._permissions:
                return self._permissions[pat]
            raise FileNotInPackageError("failed to get permissions for %s from package %s" % (file_path, pkg_path))

        self._permissions = {}

        if not content_data:
            log.debug("Calling: %s", [DPKG_BIN, "-c", pkg_path])
            content_data = subprocess.check_output([DPKG_BIN, "-c", pkg_path])
        for line in content_data.split("\n"):
            if not line:
                continue
            fields = line.split()
            if len(fields) < 6:
                raise Exception("Not enough fields: %s (%s)" % (fields, line))
            filepath = os.path.normpath("/" + fields[5])
            self._permissions[filepath] = self._parse_pkg_file_perms(line)
        if pat in self._permissions:
            return self._permissions[pat]
        raise FileNotInPackageError("failed to get permissions for %s from package %s" % (file_path, pkg_path))

    @classmethod
    def _parse_pkg_file_perms(cls, line):
        fs = FileStat()
        fs.perms = oct(cls._parse_file_mode_into_octal(line[1:10]))
        log.debug("parsed %s into '%s'" % (line[1:10], fs.perms))
        m = re.search(r'^\S+\s([\w\-]+)/([\w\-]+)\s', line)
        if not m:
            raise FileNotInPackageError("failed to parse user/group: '%s'" % line)
        (fs.user, fs.group) = m.groups()
        fields = line.split()
        if line[0] == 'l':
            (fs.typ, fs.link_to) = (FileStat.LNK, fields[7])
        elif line[0] == 'd':
            fs.typ = FileStat.DIR
        elif line[0] == '-':
            fs.typ = FileStat.REG
        elif line[0] == 'p':
            fs.typ = FileStat.FIFO
        else:
            log.warning("unexpected file type '%s' in '%s'", line[0], line)

        return fs

    @staticmethod
    def _parse_file_mode_into_octal(mode):
        octal = 0
        # rwxrwxrwx
        # Owner Read
        if mode[0] == 'r':
            octal |= 00400
        # Owner Write
        if mode[1] == 'w':
            octal |= 00200
        # Owner Execute
        if mode[2] == 'x':
            octal |= 00100
        elif mode[2] == 's':
            octal |= 0004100
        elif mode[2] == 'S':
            octal |= 0004000

        # Group Read
        if mode[3] == 'r':
            octal |= 00040
        # Owner Write
        if mode[4] == 'w':
            octal |= 00020
        # Owner Execute
        if mode[5] == 'x':
            octal |= 00010
        elif mode[5] == 's':
            octal |= 0002010
        elif mode[5] == 'S':
            octal |= 0002000

        # Other Read
        if mode[6] == 'r':
            octal |= 00004
        # Other Write
        if mode[7] == 'w':
            octal |= 00002
        # Other Execute
        if mode[8] == 'x':
            octal |= 00001
        elif mode[8] == 't':
            octal |= 0001001
        elif mode[8] == 'T':
            octal |= 0001000

        return octal

    def _get_package_path(self, pkg_name, pkg_version):
        if self._pkg_path:
            return self._pkg_path
        self._pkg_path = self._find_pkg_in_cache(pkg_name, pkg_version)
        if self._pkg_path:
            return self._pkg_path
        self._pkg_path = self._download_package(pkg_name, pkg_version)
        if self._pkg_path:
            return self._pkg_path
        raise PackageNotFoundError("Cannot find deb-archive for package %s (%s)" % (pkg_name, pkg_version))

    def _find_pkg_in_cache(self, pkg_name, pkg_version):
        for path in glob.glob(os.path.join(APT_CACHE_DIR, pkg_name+"*.deb")):
            if self._check_package_name_version(path, pkg_name, pkg_version):
                return path
        return None

    def _check_package_name_version(self, path, name, version):
        cmd = [DPKG_BIN, "-I", path]
        log.debug("Calling %s", cmd)
        output = subprocess.check_output(cmd)
        (name_ok, ver_ok) = (False, False)
        for line in output.split("\n"):
            if line == " " + self.PACKAGE_PATTERN + name:
                name_ok = True
            if line == " " + self.VERSION_PATTERN + version:
                ver_ok = True
        log.debug("name, version OK = %s, %s: %s (%s)" % (name_ok, ver_ok, name, version))
        return name_ok and ver_ok

    def _download_package(self, name, version):
        log.debug("_download_package '%s'" % name)
        if not DebFilePerms._apt_get_updated:
            check_call_with_logging([APT_GET_BIN, "update"])
            DebFilePerms._apt_get_updated = True
        cmd = [APT_GET_BIN, "install", "--reinstall", "--download-only", "-y", name+"="+version]
        check_call_with_logging(cmd)
        return self._find_pkg_in_cache(name, version)

    def package_files_list(self):
        if not self._pkg_name:
            raise Exception("Package name is not set")
        if self._file_list:
            return self._file_list

        cmd = [DPKG_BIN, '-L', self._pkg_name]
        self._file_list = []
        log.debug("Calling %s", cmd)
        content_data = subprocess.check_output(cmd)
        for line in content_data.split("\n"):
            if line and not os.path.isdir(line):
                # Skip checking the directories, because they may be implicitly added into package.
                # I.e. output may contain:
                # /lib
                # /lib/systemd
                # /lib/systemd/system
                # /lib/systemd/system/qmail.service
                # but only the last file is really present in package.
                self._file_list.append(line)
        return self._file_list

    @staticmethod
    def get_plesk_packages():
        cmd = [DPKG_QUERY_BIN, '--show', '--showformat', '${Package}\t${Maintainer}\n', '*']
        pkgs = []
        content_data = subprocess.check_output(cmd)
        for line in content_data.split("\n"):
            if line.endswith("\t" + PLESK_PACKAGER) or line.endswith("\t" + PARALLELS_PACKAGER):
                pkgs.append(line.split("\t")[0])
        return pkgs


class PermsHelper:
    _with_perms = True
    _with_owner = True
    _with_group = True
    _with_type = True

    """ Whether to ignore values defined in FILES_AC """
    _skip_file_table = False

    _pkg_helper = None

    def __init__(self, pkg=None):
        if pkg:
            self._pkg_helper = self._get_pkg_helper(pkg_name=pkg)
            self._skip_file_table = True

    def what_to_fix(self, with_perms, with_owner, with_group, with_type):
        self._with_perms = with_perms
        self._with_owner = with_owner
        self._with_group = with_group
        self._with_type = with_type

    @staticmethod
    def _get_pkg_helper(pkg_name=None):
        if RpmFilePerms.is_installed():
            return RpmFilePerms(pkg_name=pkg_name)
        elif DebFilePerms.is_installed():
            return DebFilePerms(pkg_name=pkg_name)
        else:
            raise Exception("Cannot detect package manager: no rpm or dpkg utilities found")

    def _get_file_ac(self, path):
        log.debug("PermsHelper._get_file_ac(%s)", path)
        if not self._skip_file_table and path in FILES_AC:
            if FILES_AC[path] is None:
                log.info("File %s does not exist on current OS.", path)
                return FileStat()
            perms = FILES_AC[path][0]
            owner = FILES_AC[path][1]
            group = FILES_AC[path][2]
            typ = FILES_AC[path][3]
            link = None
            if typ == FileStat.LNK:
                link = FILES_AC[path][4]
            st = FileStat(perms=perms, user=owner, group=group, typ=typ, link_to=link)
            log.debug("return %s", st)
            return st

        return self._get_package_file_perms(path)

    def _get_package_file_perms(self, path):
        log.debug("PermsHelper._get_package_file_perms(%s)", path)
        pkg_helper = self._pkg_helper
        if not pkg_helper:
            pkg_helper = self._get_pkg_helper()
        return pkg_helper.get_file_perms(path)

    def fix_file(self, path, exp=None):
        if not exp:
            exp = self._get_file_ac(path)

        if exp == FileStat():
            if os.path.exists(path) or os.path.islink(path):
                raise FsError("The file %s is present but it should not be." % path)
            return

        # Fixing type should go first:
        if self._with_type:
            if exp.typ == FileStat.LNK:
                self._fix_symlink(exp.link_to, path)
            elif exp.typ == FileStat.DIR and not os.path.exists(path):
                self._info("Creating non-existing directory '%s'" % path)
                os.mkdir(path, 0700)

            act = FileStat(mode=os.lstat(path).st_mode)
            if exp.typ != act.typ:
                raise Exception("Expected and actual types of file '%s' do not match: %s != %s ."
                                % (path, exp.typ, act.typ))

        if not os.path.exists(path) and not os.path.islink(path):
            raise FsError("The file %s was not found." % path)

        path_stat = os.lstat(path)

        # fixing owner
        if self._with_owner:
            (uids, owners) = get_owners(exp.user)
            if not uids or not owners:
                log.warning("Expected users do not exist: %s. Skipping.", exp.user)
            elif path_stat.st_uid not in uids:
                self._info("Changing owner on %s to %s (%s)" % (path, owners[0], uids[0]))
                os.lchown(path, uids[0], -1)

        # fixing group
        if self._with_group:
            (gids, groups) = get_groups(exp.group)
            if not gids or not groups:
                log.warning("Expected groups do not exist: %s. Skipping", exp.group)
            elif path_stat.st_gid not in gids:
                self._info("Changing group on %s to %s (%s)" % (path, groups[0], gids[0]))
                os.lchown(path, -1, gids[0])

        # permissions may be changed after setting owner/group
        path_stat = os.lstat(path)

        # fixing permissions
        if self._with_perms:
            if not os.path.islink(path):
                perms = exp.perms.split(',')
                log.debug("exp perms %s = %s, act mode = %s" % (path, perms, oct(path_stat.st_mode & 07777)))
                if oct(path_stat.st_mode & 07777) not in perms:
                    self._info("Changing permissions on %s to %s" % (path, perms[0]))
                    os.chmod(path, int(perms[0], 8))
            else:
                log.debug("Changing permissions on symlinks is not possible: %s" % path)

    @classmethod
    def _fix_symlink(cls, src, dst):
        if os.path.exists(dst) and not os.path.islink(dst):
            raise Exception("Cannot fix %s: file exists and it is not a symlink")
        if os.path.islink(dst):
            target = os.readlink(dst)
            if target == src:
                return
            os.unlink(dst)
        cls._info("Recreating symlink %s -> %s" % (dst, src))
        os.symlink(src, dst)

    def fix_package(self):
        if not self._pkg_helper:
            raise Exception("Package helper is not instantiated")
        files = self._pkg_helper.package_files_list()
        for f in files:
            try:
                self.fix_file(f)
            except FileNotInPackageError as e:
                log.warning(e)

    def check_file(self, path, exp=None):
        if not exp:
            exp = self._get_file_ac(path)

        if exp == FileStat():
            if os.path.exists(path) or os.path.islink(path):
                raise FsError("The file %s is present but it should not be." % path)
            return

        if not os.path.exists(path) and not os.path.islink(path):
            raise FsError("The file %s was not found." % path)

        log.debug("get_file_ac returned %s" % exp)

        st = os.lstat(path)
        errors = [
            validate_perms(path, st, exp.perms),
            validate_owner(path, st, exp.user),
            validate_group(path, st, exp.group),
            validate_type(path, st, exp.typ, exp.link_to),
        ]
        if any(errors):
            raise FsError("\n".join(filter(None, errors)))

    def check_package(self):
        if not self._pkg_helper:
            raise Exception("Package helper is not instantiated")
        files = self._pkg_helper.package_files_list()
        error_lines = []
        for f in files:
            try:
                self.check_file(f)
            except FileNotInPackageError as e:
                log.warning(e)
            except FsError as e:
                error_lines += str(e).split("\n")
        if error_lines:
            raise FsError("\n".join(error_lines))

    @classmethod
    def get_plesk_packages(cls):
        return cls._get_pkg_helper().get_plesk_packages()

    @staticmethod
    def _info(msg):
        print(msg)

# vim: ts=4 sts=4 sw=4 et :
