# Copyright 1999-2017. Plesk International GmbH. All rights reserved.

import subprocess
import os
import sys
import re
from contextlib import contextmanager
from distutils.version import StrictVersion
import logging as log

import apt_pkg

from os_detect import OS
import plesk_constants
import pum_errors


CODENAMES = {
    "Debian": {
        "7": "wheezy",
        "8": "jessie",
        "9": "stretch",
    },
    "Ubuntu": {
        "12.04": "precise",
        "14.04": "trusty",
        "16.04": "xenial",
        "18.04": "bionic",
    }
}

PLESK_VENDOR = "Plesk"
PLESK_PRODUCT = "Plesk"


class AptWrapper:
    _stable_mark = None

    def __init__(self, cache, skip_update=False, unsafe=False):
        self.cache = cache
        self._unsafe = unsafe
        if not skip_update and cache is not None:
            self.cache.update()
            self.cache.open()
        self.pm = "apt"
        self._exclude = []

        # force non-interactive mode
        os.environ["DEBIAN_FRONTEND"] = "noninteractive"
        apt_pkg.init()
        apt_pkg.config["Dpkg::Options::"] = "--force-confdef"
        apt_pkg.config["Dpkg::Options::"] = "--force-confold"

        # apt OS specific parameters
        self._vendor = OS.distname()  # "Debian" or "Ubuntu"
        if OS.is_debian():
            stable = self._get_debian_stable_name(skip_update)
            stable = stable + "|" if stable is not None else ""
            self._archive_re = r'^(' + stable + CODENAMES["Debian"][OS.version()] + r')$'
        elif OS.is_ubuntu():
            self._archive_re = r'^' + CODENAMES["Ubuntu"][OS.version()] + r'.*'
        self._archive = re.compile(self._archive_re)

    @staticmethod
    def _get_package_info(package, version):
        return {
            "name": package.name,
            "arch": version.architecture,
            "full_version": version.version
        }

    @staticmethod
    def _get_package_source(version):
        origin = version.origins[0]
        return "%s %s/%s" % (origin.site, origin.archive, origin.component)

    def get_installed_packages(self):
        """
        Retrieve the list of the installed packages.

        @return dictionary of (package name, package info), where package info
                is extended with source id and summary
        """
        installed_packages = {}

        for package in self.cache:
            if not package.is_installed:
                continue

            version = package.installed

            package_info = self._get_package_info(package, version)
            package_info["summary"] = package.installed.summary
            package_info["repo_from"] = self._format_repo_origin_info(package.installed.origins[0])

            installed_packages[package.name] = package_info

        return installed_packages

    @staticmethod
    def _is_vendor_repo_label(label):
        if OS.is_debian():
            return label in ("Debian", "Debian-Security")
        elif OS.is_ubuntu():
            return label == "Ubuntu"
        log.warning("Unknown OS: %s", OS.distname())
        return False

    @staticmethod
    def _check_vendor_labels_compat(old, new):
        return (old == new or
                (AptWrapper._is_vendor_repo_label(old) and AptWrapper._is_vendor_repo_label(new)))

    @staticmethod
    def _check_ubuntu_archives_compat(old, new):
        codename = CODENAMES["Ubuntu"][OS.version()]
        ubuntu_archives = (codename, codename + "-updates", codename + "-security")
        return old in ubuntu_archives and new in ubuntu_archives

    @classmethod
    def check_archives_compatibility(cls, old, new, skip_update=False):
        if old == new:
            return True
        if ((OS.is_ubuntu() and
             cls._check_ubuntu_archives_compat(old, new))):
            return True
        if OS.is_debian():
            this_codename = CODENAMES["Debian"][OS.version()]
            this_stable_name = cls._get_debian_stable_name(skip_update=skip_update)
            if ((old in (this_codename, this_stable_name) and
                 new in (this_codename, this_stable_name))):
                return True
        return False

    @staticmethod
    def _check_plesk_labels_compat(old_label, new_label):
        if old_label == new_label:
            return True
        if not old_label:
            # upgrade from 'now' package
            oldver = StrictVersion("0.0.0")
        else:
            if not old_label.startswith(PLESK_PRODUCT):
                return False
            old_match = re.search(r'\d+\.\d+\.\d+', old_label)
            if not old_match:
                return False
            oldver = StrictVersion(old_match.group(0))

        if not new_label.startswith(PLESK_PRODUCT):
            # upgrade from now-package to any package from Plesk repo is allowed
            return not old_label
        new_match = re.search(r'\d+\.\d+\.\d+', new_label)
        if not new_match:
            return False
        newver = StrictVersion(new_match.group(0))
        curver = StrictVersion(plesk_constants.PRODVERSION)
        return oldver <= newver and newver <= curver

    def check_origins_compatibility(self, old, new, skip_update=False):
        now_package = old.archive == "now" and not old.origin and not old.label

        # check archives:
        if not now_package and not self.check_archives_compatibility(old.archive, new.archive, skip_update=skip_update):
            log.debug("origin archive mismatch: %s != %s", old.archive, new.archive)
            return False
        if now_package and not self._archive.match(new.archive):
            log.debug("origin archive mismatch: %s !~ %s", old.archive, self._archive_re)
            return False

        # check now package:
        if now_package:
            if new.origin == self._vendor and self._is_vendor_repo_label(new.label):
                return True
            if new.origin == PLESK_VENDOR and self._check_plesk_labels_compat(None, new.label):
                return True
            log.debug("origin mismatch for 'now' package: %s by %s != %s by %s", old.label, old.origin, new.label, new.origin)
            return False

        # check origins:
        if old.origin != new.origin:
            log.debug("origin origin mismatch: %s != %s", old.origin, new.origin)
            return False

        if old.origin == self._vendor:
            if not self._check_vendor_labels_compat(old.label, new.label):
                log.debug("vendor origin label mismatch: %s != %s", old.label, new.label)
                return False
        elif old.origin == PLESK_VENDOR:
            if not self._check_plesk_labels_compat(old.label, new.label):
                log.debug("plesk origin label mismatch: %s != %s", old.label, new.label)
                return False
        elif old.label != new.label:
            log.debug("origin label mismatch: %s != %s", old.label, new.label)
            return False

        # everything is passed :)
        return True

    def mark_proper_upgrade(self, package):
        installed_origin = package.installed.origins[0]

        for v in package.versions:
            if ((v.version != package.installed.version and
                 self.check_origins_compatibility(installed_origin, v.origins[0]))):
                package.candidate = v
                package.mark_upgrade()
                return v
        return None

    def _prepare_update(self, packages):
        if not packages:
            self.cache.upgrade(dist_upgrade=self._unsafe)
            pkgs = [x.name for x in self.cache.get_changes()]
            self.cache.clear()
        else:
            pkgs = packages

        added = []
        for p in pkgs:
            if p not in self.cache:
                log.warning("package '%s' not found", p)
                continue

            package = self.cache[p]
            if not package.is_installed:
                log.warning("package '%s' not installed", p)
                continue

            if not package.candidate:
                log.debug("no update candidate for '%s'", p)

            if self._unsafe:
                candidate = package.candidate
                package.mark_upgrade()
            else:
                candidate = self.mark_proper_upgrade(package)
                if not candidate:
                    log.debug("proper candidate for '%s' not found", p)
                    continue
                else:
                    log.debug("proper candidate for '%s' found: %s", p, candidate)

            # rollback transaction if any packages are in exlude list or marked for delete:
            if any([(ch.name in self._exclude or ch.marked_delete) for ch in self.cache.get_changes()]):
                self.cache.clear()
                for (name, cand) in added:
                    self.cache[name].candidate = cand
                    self.cache[name].mark_upgrade()
            else:
                added.append((p, candidate))

    def exclude_packages(self, packages):
        self._exclude = packages

    @staticmethod
    def _format_repo_origin_info(repo_origin):
        now_package = repo_origin.archive == "now" and not repo_origin.origin and not repo_origin.label
        if now_package:
            return "now"
        # It should produce descriptions like "Debian for oldstable by Debian", "Plesk 17.0.11 packages for wheezy by Plesk"
        return "%s for %s by %s" % (repo_origin.label, repo_origin.archive, repo_origin.origin)

    def simulate_update(self, packages):
        """
        Perform a simulation of update of the given packages.

        @param packages list of packages (names) to simulate update of
        @return dictionary of (package name, updated package info) -- value type will be refined
        """
        transaction_info = {}

        self._prepare_update(packages)

        for package in self.cache.get_changes():
            item = self._get_package_info(package, package.candidate)
            if package.marked_delete:
                item["full_version"] = None
            if package.installed is not None:
                item["from"] = package.installed.version
                item["summary"] = package.installed.summary
                item["repo_from"] = self._format_repo_origin_info(package.installed.origins[0])
                item["repo_to"] = self._format_repo_origin_info(package.candidate.origins[0])
            else:
                item["from"] = None
                item["summary"] = package.candidate.summary
                item["repo_from"] = None
                item["repo_to"] = self._format_repo_origin_info(package.candidate.origins[0])
            transaction_info[package.name] = item

        return transaction_info

    def update_packages(self, packages):
        """
        Install the newest versions of the given packages.

        @param packages list of packages (names) to be updated
        """
        self._prepare_update(packages)

        pkgs = []
        for package in self.cache.get_changes():
            pkgs.append(package.name)
        if len(pkgs) > 0:
            log.info("updating packages: %s" % (" ".join(pkgs)))
        else:
            log.info("no packages to update")

        self.cache.commit()

    def print_package_details(self, package):
        """
        Display detailed package information.

        @param package name of the installed package to display information on
        """
        if package in self.cache and self.cache[package].is_installed:
            subprocess.check_call(["dpkg", "-s", package])
        else:
            # Error message similar to yum one
            raise Exception("package %s is not installed" % package)

    @classmethod
    def _get_debian_stable_name(cls, skip_update=False):
        if cls._stable_mark is not None:
            return cls._stable_mark
        if not skip_update:
            with open(os.devnull, "wb") as devnull:
                subprocess.check_call(["apt-get", "update"], stdout=devnull)
        p = subprocess.Popen(["apt-cache", "policy"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        if p.returncode != 0:
            log.warning("apt-cache policy failed (%s): %s", p.returncode, err)
            return None
        re_main_str = r"""^.*a=(\w+),n=%s,l=Debian,c=main""" % CODENAMES["Debian"][OS.version()]
        log.debug("re_main_str = %s", re_main_str)
        re_main = re.compile(re_main_str)
        for line in out.split("\n"):
            matches = re_main.match(line)
            if matches is not None:
                cls._stable_mark = matches.group(1)
                return cls._stable_mark
        log.warning("Unable to detect archive for debian release:\n%s" % out)
        return None

    def get_compatible_repos_list(self):
        repos_map = {}
        for package in self.cache:
            if package.installed:
                repos_map[self._format_repo_origin_info(package.installed.origins[0])] = package.installed.origins[0]
            for v in package.versions:
                repos_map[self._format_repo_origin_info(v.origins[0])] = v.origins[0]

        log.debug("list of repositories: %s", repos_map.keys())
        res = []
        for r1 in repos_map.values():
            for r2 in repos_map.values():
                res.append((self._format_repo_origin_info(r1),
                            self._format_repo_origin_info(r2),
                            self.check_origins_compatibility(r1, r2)))
        return res

    def check(self, full):
        """ Performs  database  sanity  and  consistency checks for all packages. """
        p = subprocess.Popen(["dpkg", "--audit"], stdout=subprocess.PIPE)
        (out, err) = p.communicate()
        if out != "":
            print out,
        rc = 0
        if p.returncode != 0:
            rc = p.returncode
        elif out != "":
            rc = 1
        sys.exit(rc)

    @contextmanager
    def locking(self):
        lockfile = apt_pkg.config.find_dir("Dir::Cache::Archives") + "lock"
        lock = apt_pkg.get_lock(lockfile)
        if lock < 0:
            raise pum_errors.PMLockError("Failed to lock %s" % lockfile)

        try:
            yield
        finally:
            os.close(lock)

# vim: ts=4 sts=4 sw=4 et :
