# Copyright 1999-2017. Plesk International GmbH. All rights reserved.

import plesk_log
import plesk_constants
from optparse import OptionParser
import sys, json, os, subprocess, copy
from plesk_interproc_locks import InterProcessLock
import plesk_service

RS_DB_NAME = '/var/lib/plesk/register_service/data.db'
RS_LOCK_NAME = 'register_service'

CAPABILITY = 'capabilities'
RESOURCE_CONTROLLER = 'with-resource-controller'
OLD_RESOURCE_CONTROLLER = 'resource-controller'
NO_WATCHDOG = 'no-watchdog'
STD_CAPABILITIES = frozenset([RESOURCE_CONTROLLER, NO_WATCHDOG])

SERVICES = 'services'
SERVICE_NAME = 'name'
SERVICE_IS_RUNNING = 'is_running'

RESCTRL_REGISTER_CMD = '--register-service'
RESCTRL_UNREGISTER_CMD = '--unregister-service'


def db_lock():
    return InterProcessLock(RS_LOCK_NAME)


def get_data():
    try:
        with open(RS_DB_NAME, 'r') as db:
            data = json.load(db)
            for name, properties in data.iteritems():
                capabilities = properties[CAPABILITY]
                if OLD_RESOURCE_CONTROLLER in capabilities:
                    capabilities[RESOURCE_CONTROLLER] = capabilities[OLD_RESOURCE_CONTROLLER]
                    del capabilities[OLD_RESOURCE_CONTROLLER]
    except:
        log.debug('Fail to parse data from %s', RS_DB_NAME)
        data = {}
    return data


def set_data(data):
    if not os.path.isfile(RS_DB_NAME):
        log.info('Create DB %s', RS_DB_NAME)

    with open(RS_DB_NAME, 'w') as db:
        log.info('Write data to %s', RS_DB_NAME)
        json.dump(data, db)


def normalize_capabilities(options, check_on_empty=False):
    capabilities = set()

    for cap in STD_CAPABILITIES:
        if getattr(options, cap.replace('-', '_')):
            capabilities.add(cap)

    if options.capabilities:
        for cap in options.capabilities.split(','):
            capabilities.add(cap.lower().strip())
        for cap in capabilities.difference(STD_CAPABILITIES):
            raise ValueError('Unknown capability %s' % cap)

    options.capabilities = list(capabilities)

    if check_on_empty and len(capabilities) == 0:
        raise ValueError('You should specify capabilities!')


def get_service(data, name):
    if name not in data:
        log.debug('Service %s doesn`t exist', name)
        return None
    return data[name]


def call_resctrl(action, name):
    RESCTRL = os.path.join(plesk_constants.PRODUCT_ROOT_D, 'admin/sbin/resctrl')

    if os.path.isfile(RESCTRL) and os.access(RESCTRL, os.X_OK):
        return subprocess.check_call([RESCTRL, action, name])


def set_service_impl(name, service, options):
    capabilities = service[CAPABILITY]
    for cap in options.capabilities:
        if cap not in capabilities:
            log.info('Add capability %s to the service %s', name, cap)
            capabilities[cap] = True
            if cap == RESOURCE_CONTROLLER:
                call_resctrl(RESCTRL_REGISTER_CMD, name)
        else:
            log.debug('Service %s already has capability %s', cap, name)


def add_service(name, options):
    normalize_capabilities(options)

    with db_lock():
        data = get_data()
        orig_data = copy.deepcopy(data)
        service = get_service(data, name)
        if not service:
            log.info('Add new service %s', name)
            service = data.setdefault(name, {})
            service[CAPABILITY] = {}
        else:
            log.info('Service %s already exists', name)

        set_service_impl(name, service, options)

        if orig_data != data:
            set_data(data)


def rm_service(name, options):
    with db_lock():
        data = get_data()
        service = get_service(data, name)
        if service:
            log.info('Remove service %s', name)
            if service[CAPABILITY].get(RESOURCE_CONTROLLER, False):
                call_resctrl(RESCTRL_UNREGISTER_CMD, name)
            del data[name]
            set_data(data)


def unset_service(name, options):
    normalize_capabilities(options, True)

    with db_lock():
        data = get_data()
        service = get_service(data, name)
        if not service:
            return

        orig_data = copy.deepcopy(data)
        capabilities = service[CAPABILITY]
        for cap in options.capabilities:
            if cap not in capabilities:
                log.debug('Service %s doesn`t have capability %s', name, cap)
            else:
                log.info('Remove capability %s from the service %s', cap, name)
                del capabilities[cap]
                if cap == RESOURCE_CONTROLLER:
                    call_resctrl(RESCTRL_UNREGISTER_CMD, name)

        if data != orig_data:
            set_data(data)


def set_service(name, options):
    normalize_capabilities(options, True)

    with db_lock():
        data = get_data()
        service = get_service(data, name)
        if not service:
            return
        orig_data = copy.deepcopy(data)
        set_service_impl(name, service, options)
        if data != orig_data:
            set_data(data)


def list_services(options):
    with db_lock():
        data = get_data()

    services = []
    for name, properties in data.iteritems():
        service = {SERVICE_NAME: name}
        services.append(service)
        service_capabilities = properties[CAPABILITY]
        for capability in STD_CAPABILITIES:
            if capability in service_capabilities:
                service[capability.replace('-', '_')] = True
        if options.full:
            try:
                if plesk_service.action(name, 'status'):
                    service[SERVICE_IS_RUNNING] = True
            except Exception as ex:
                log.error('%s', ex)
                log.debug(ex, exc_info=True)

    out_data = {SERVICES: services}
    print json.dumps(out_data)


def do_main():
    plesk_log.setDefaultLevel('warning')

    global log
    log = plesk_log.getLogger(__name__)

    parser = OptionParser(
        usage='Usage: %prog {'
        '--add-service <service_name>|'
        '--rm-service <service_name>|'
        '--set-service <service_name>|'
        '--unset-service <service_name>|'
        '--list-services} [options]')

    parser.add_option(
        '-v', '--verbose', action='count',
        help='Increase verbosity (can be specified multiple times).')

    parser.add_option('-a', '--add-service', help='Add service to database')
    parser.add_option('-r', '--rm-service', help='Remove service from database')
    parser.add_option('-s', '--set-service', help='Add service capabilities')
    parser.add_option(
        '-u', '--unset-service', help='Remove service capabilities')

    parser.add_option(
        '-l', '--list-services', action='store_true', default=False,
        help='Show list of services in database')

    parser.add_option(
        '-c', '--capabilities', type='string', default='',
        help='Comma separated list of capabilities')

    for cap in STD_CAPABILITIES:
        parser.add_option(
            '--' + cap, action='store_true', default=False,
            help=('Add capabitlity "%s"' % cap))

    parser.add_option(
        '-f', '--full', action='store_true', default=False,
        help='Show full information about services')

    (options, args) = parser.parse_args()

    if options.verbose == 1:
        plesk_log.setDefaultLevel('info')
    elif options.verbose >= 2:
        plesk_log.setDefaultLevel('debug')
    log = plesk_log.getLogger(__name__)

    if options.add_service:
        add_service(options.add_service, options)
    elif options.rm_service:
        rm_service(options.rm_service, options)
    elif options.set_service:
        set_service(options.set_service, options)
    elif options.unset_service:
        unset_service(options.unset_service, options)
    elif options.list_services:
        list_services(options)
    else:
        parser.error(
            'Action is not specified. '
            'Please, select one of: --add-service, --rm-service, '
            '--set-service, --unset-service, --list-service')


def main():
    try:
        exit_code = do_main()
    except Exception as ex:
        log.error('%s', ex)
        log.debug(ex, exc_info=True)
        exit_code = 1

    sys.exit(exit_code)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
